/******************************************************************
 * ep-approx-list.c
 * - ߎ}CjOASYD
 *   ɁCҏW{insert, delete, exchange}gD
 * - A frequent sequence miner for approximate string patterns
 *   where {insert, delete, exchange}-operations are allowed
 *
 * Modified on 25 FEB 2004 by arim
 * - ̐炷߂ -l IvViŒҏWjgݍ񂾁D
 *   l = 0 ̏ꍇ́Cp^[̓eLXg̕ɐ̂ŁC
 *   n^2̉ȂDʂɁC𐔂͍Xn^{l+2}łD
 * Modified on 14 FEB 2004 by arim
 * - R}hCƃIvVŁCp[^ǂݍ߂悤ɂD
 *   requires GNU getopt library (contained in gcc package).
 * Created on 14 FEB 2004 by arim@i.kyushu-u.ac.jp
 *   based on ep-approx-list-all2.c 
 *   ep-approx-list-all2.c: 
 *   Created on 4 FEB 2004 by {uchida,arim}@i.kyushu-u.ac.jp.
 *   based on ep-approx-list-all.c 
 * $Id: main.c,v 1.6 2004/02/25 20:31:59 arim Exp $
 ******************************************************************/

#include "ap-util.h"
#include "ap-carray.h"
#include "ap-iarray.h"
#include <assert.h>
#include <stdlib.h>
#include <time.h>
#include <getopt.h> /* for gnu-dependent getopt-long library */
#include "ep-mine-api.h"

/******************************************************************
 * Macros
 ******************************************************************/

/* For gnuopt
 */
#define EOFOPT (-1) /* End of option */

/* General variables
 */

int UNDEF = -1;
int APP_MINSUP_RELATIVE = 1.0; 
int MAXINS = 1;
int MAXDEL = 1;
int MAXEX  = 1;
int DELTA_WIN  = 0; //default
int ENABLE_PRINT_PATTERN  = TRUE;
int WINDOW_FREE = FALSE;

/* General variables*/
char filename[128];
int   app_minsup = UNDEF; 	/* minimum support in */
float app_minsup_percent = UNDEF; 	/* minimum support in percent */
int   maxscore = 0;                /* maximum edit distance */
int   minscore = INTMAX;           /* minimum edit distance */

/******************************************************************
 * Global
 ******************************************************************/

//int GAMMA[20][20];

/* Usage: vO̎g */
char USAGE_STRING[] = 
"*** epi-approx.exe  - a program for finding approximate subsequences *** \n" 
"*** Paper: Y. Uchida, T. Asai, H. Arimura, DEWS2004, I-7-04, 2004.   *** \n" 
"*** 2004 - 2006, Hiroki Arimura, Yuzo Uchida, All right reserved.    *** \n" 
"\nUSAGE: \n" 
" main.exe -k score -s minsup OPTIONS < file\n"
" find all approx. subsequences appearing in at least minsup input  \n"
" sequences in which the pattern occurs within edit distance score. \n" 
"\nEXAMPLE\n" 
" main < input.txt #find approximate motifs in input.txt"
" main -k 2 -s 100 < input.txt #find approx. motifs with max edit distance 2 and absolute minsup 100 (10%)\n"
" main -k 2 -r 0.1 < input.txt #find approx. motifs with max edit distance 2 and relative minsup 0.1 (10%)\n"
" main -n < input.txt #print only the pattern histogram\n"
"\n OPTIONS\n" 
" -k score    \t maximum integer score allowed (default 0)\n"
" -s num    \t minimum support num (default the number of input strings)\n"
"           \t (default: the number of input strings)\n"
" -r ratio  \t relative minimum support 0 <= ratio <= 1 (default 1.0)\n"
" -p        \t printing patterns (default)\n"
" -n        \t not printing patterns\n"
"(-i|-d|-e) s\t Set the edit score of the opr i/d/e to s (default 1)\n"
"(-I|-D|-E) \t Disable the specified edit operation (score := INT_MAX)\n"
" -w width  \t Using windowed epsode with width k (experimental)\n" 
" -f        \t Using window-free epsode (experimental)\n" 
" -l score  \t At least one string must contain the pattern in score (array only) (experimental)\n"
" Note: Edit operations: i/INSERT, d/DELETE/,  e/x/EXCHANGE\n" 
; /* End of USAGE_STRING */


/******************************************************************
 * Main
 ******************************************************************/

void printAnOption(char *str, int optind, char *optarg) {
#ifdef DEBUG
  printf("#%s: optind=%d optarg=[%s]\n", str, optind, optarg);
#else
  ;
#endif
}


/* IvV̏w */
static char* OptionFormat = "hwIDEXi:d:e:x:W:s:r:k:l:pnf"; 

void readOptions(int argc, char *argv[]) {
  /* Option variables */
  int C; /* Option letter */
  extern int optind; 
  extern char *optarg;

  //  if (argc <= 1) { /* no arguments */
  //    printf(USAGE_STRING);
  //    exit(0);
  //  }

  while ((C = getopt(argc, argv, OptionFormat)) != EOFOPT) {
#ifdef DEBUG
    printf("#getopt: C=%d\n", C);
#endif
    switch (C) {
    case 's':
      /* -s app_minsup: the number of input sequences as minimum support */
      printAnOption("s", optind, optarg);
      app_minsup = atoi(optarg);
      break;
    case 'r':
      /* -r app_minsup: the percent as minimum support */
      printAnOption("r", optind, optarg);
      app_minsup_percent = atof(optarg);
      break;
    case 'k':
      /* -k maxscore: the percent as minimum support */
      printAnOption("k", optind, optarg);
      maxscore = atoi(optarg);
      break;
      //Flag for edit operations
    case 'l':
      /* -l minscore: the percent as minimum support */
      printAnOption("l", optind, optarg);
      minscore = atoi(optarg);
      break;
      //Flag for edit operations
    case 'I':
      printAnOption("I", optind, optarg);
      //USEINS = TRUE;
      MAXINS = INTMAX;
      break;
    case 'D':
      printAnOption("D", optind, optarg);
      //USEDEL = TRUE;
      MAXDEL = INTMAX;
      break;
    case 'E':
    case 'X':
      printAnOption("E", optind, optarg);
      //USEEX = TRUE;
      MAXEX  = INTMAX;
      break;
      //Set non-standard edit score
    case 'w':
      printAnOption("W", optind, optarg);
      //USEINS = TRUE;
      MAXINS = 1;
      MAXDEL = INTMAX;
      MAXEX  = INTMAX;
      DELTA_WIN = 1;
      break;
    case 'f':
      printAnOption("F", optind, optarg);
      WINDOW_FREE = TRUE;
      maxscore = INTMAX;
      break;
    case 'i':
      printAnOption("i", optind, optarg);
      //USEINS = TRUE;
      MAXINS = atoi(optarg);
      break;
    case 'd':
      printAnOption("d", optind, optarg);
      //USEDEL = TRUE;
      MAXDEL = atoi(optarg);
      break;
    case 'e':
    case 'x':
      printAnOption("e", optind, optarg);
      //USEEX = TRUE;
      MAXEX = atoi(optarg);
      break;
    case 'p':
      printAnOption("p", optind, optarg);
      ENABLE_PRINT_PATTERN = TRUE;
      break;
    case 'n':
      printAnOption("n", optind, optarg);
      ENABLE_PRINT_PATTERN = FALSE;
      break;
    case 'h':
      printAnOption("h", optind, optarg);
      printf(USAGE_STRING);
      exit(0);
    default:
      printAnOption("*", optind, optarg);
      printf("#getopt returned char: %d\n", C);
      exit(0);
    }
  }
}

void readArgs(int argc, char *argv[]) {
  int  _argc   = 0;
  char **_argv = NULL;
  int i;
  char *str;
  _argc = argc - optind;
  _argv = argv + optind;

  for (i = 0; i < _argc; i++) {
    /* Processing the next argument */
    str = _argv[i];
    printf("#argv[%d]=[%s]\n", i, str);
  }
}

int main(int argc, char *argv[])
{
  Database database;
  Eventset eventset;
  clock_t gstime, gtime;
	int   i; /* working */

	printf("### argc=%d\n", argc);

#ifdef DEBUG
  printf("#Processing options...\n");
#endif 

	readOptions(argc, argv);

#ifdef DEBUG
	printf("#Processing non-option arguments...\n");
#endif
	readArgs(argc, argv);

	/*
	 * Body
	 */
	gstime=clock();

	init_pattern_stat();

	printf("Loading sequences...\n");
	database = db_read();
	printf("#database size:\t%d (seqs)\t%d (bytes)\n",
	       db_get_num_seqs(database), db_get_num_chars(database));
	//db_print(database);

	//Setting MINSUP
	if (app_minsup == UNDEF) {
	  if (app_minsup_percent == UNDEF) {
		app_minsup = db_get_num_seqs(database);
	  }
	  else {
		app_minsup = (int) (db_get_num_seqs(database)*app_minsup_percent);
	  }
	}
	if (app_minsup <= 0) {
	  printf("error: app_minsup should be >= 1: %d\n", app_minsup);
	  exit(1);
	}

	eventset = make_es(database);
	es_print(eventset);

	printf("Printing parameters...\n");
	printf("#param: \tapp_minsup=%d \tapp_minsuppct=%f \n", app_minsup, app_minsup_percent);
	printf("#param: \tmaxscore=%d \tminscore=%d \n", maxscore, minscore);
	printf("#param: \tMAXINS=%d \tMAXDEL=%d \tMAXEX=%d\n",
	       MAXINS, MAXDEL, MAXEX);

	//printf("Making gamma that event exchange matrix");
	//gamma = make_gamma(eventset);

	printf("Mining patterns...\n");
	episode_print_header();
	episode_init(); 
	episode_mine(eventset, database, app_minsup, maxscore);

	//free(line);

	//Print the histogram
	print_pattern_stat();

	/* Reporting statistics
	 */
	gtime=clock()-gstime; 

	printf("#database size:\t%d (seqs)\t%d (bytes)\n",
	       db_get_num_seqs(database), db_get_num_chars(database));
	printf("#patterns num:\t%d (patterns)\n", pattern_id);
	printf("#totalTime:   \t%.9f (sec)\t%d (min)\n",
	       (double)gtime/CLK_TCK, (int)gtime/CLK_TCK/60);

	printf("done.\n");
	exit(0);
}
